#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

import math,re,os
#valid_nt=['A','C','G','T','U','Y','R','W','S','K','M','B','D','H','V','N']
valid_nt = 'ACGTUYRWSKMBDHVN'
IUPAC = 'ACMGRSVTWYHKDBN'

comp_dct={'a':'t','c':'g','g':'c','t':'a','b':'v','h':'d','k':'m','m':'k','n':'n','r':'y','s':'s','u':'a','v':'b','w':'w','y':'r',' ':' '}

degenerated_dct={'a':'a','c':'c','g':'g','t':'t',
                 'ac':'v','h':'d','k':'m','m':'k',
                 'n':'n','r':'y','s':'s','u':'a',
                 'v':'b','w':'w','y':'r',' ':' '}

def blast2seq(query_name,seq1,subject_name,seq2):
    """launch bl2seq analysis and retrieve results"""
    import tempfile
    fn1=tempfile.mktemp()
    f1=open(fn1,'w').write(seq1)

    fn2=tempfile.mktemp()
    f2=open(fn2,'w').write(seq2)

    fn3=tempfile.mktemp()

    import commands

    r=commands.getoutput('bl2seq -p blastn -i %s -j %s -o %s' % (fn1,fn2,fn3))
    if r:
        return 'bl2seq error: '+str(r)

    os.remove(fn1)
    os.remove(fn2)

    f=open(fn3)
    rows=f.readlines()
    f.close()

    #print fn3
    os.remove(fn3)

    query,subject,match,identity='','','',''
    query_start,query_end,subject_start,subject_end=0,0,0,0

    flag=False
    result=[]
    for row in rows:
        
        if row.find('***** No hits found ******')<>-1:
            return 'no hits found'

        if row.find(' Score = ')==0:
            if flag:
                result.append({'query_name':query_name,'subject_name':subject_name,'query':query,'match':match,'subject':subject,'identity':identity,'query_start':query_start,'query_end':query_end,'subject_start':subject_start,'subject_end':subject_end})
                query,subject,match='','',''
                query_start,query_end,subject_start,subject_end=0,0,0,0
            else:
                flag=True


        if row.find(' Identities = ')==0:
            identity=row.strip().split('(')[1].split(')')[0]

        if row.find('Query:')==0:
            r=row.strip().split(' ')
            while '' in r:
                r.remove('')
            if not query_start:
                query_start=r[1]
            query_end=r[3]
            query=query+r[2]
            match=match+rows[rows.index(row)+1][11:].strip()
           
           
        if row.find('Sbjct:')==0:
            r=row.strip().split(' ')
            while '' in r:
                r.remove('')
            subject=subject+r[2]
            if not subject_start:
                subject_start=r[1]
            subject_end=r[3]
          
    result.append({'query_name':query_name,'subject_name':subject_name,'query':query,'match':match,'subject':subject,'identity':identity,'query_start':query_start,'query_end':query_end,'subject_start':subject_start,'subject_end':subject_end})

    return result

def oligo_length(seq):
    '''Determine the nucleotide sequence's length'''
    count=0
    seq=str(seq)
    for nt in seq:
        if nt.upper() in IUPAC:
            count+=1 
    return count

def complement_nt(nt):
    '''Complement a nucleotide'''
    nt = str(nt)
    if comp_dct.has_key(nt.lower()):
        if nt.islower():
            return comp_dct[nt]
        else:
            return comp_dct[nt.lower()].upper()
    else:
        return ''

def convert2IUPAC(seq):
    """Convert sequence to IUPAC sequence
    return str type
    """
    regexp = '\([^\(^\)]{1,}\)'
    seq = str(seq)
    seq = seq.upper()
    while seq.count('(') and seq.count(')'):
        r=re.search(regexp,seq).group()
        r=r.replace('(','').replace(')','')
        v=('A' in r)+('C' in r)*2+('G' in r)*4+('T' in r)*8
        seq=seq.replace('('+r+')',IUPAC[v-1])
    return seq

def rev_comp_seq(seq):
    """return complemented and reverted sequence"""
    revcomp_s=''
    for nt in seq:
        revcomp_s=complement_nt(nt)+revcomp_s
    return revcomp_s    

def ndegen(seq):
    """return number of different not degenerated sequences"""
    nd=1
    for nt in seq:
        if nt in ['M','R','S','W','Y','K']:
            nd=nd*2
        if nt in ['V','B','D','H']:
            nd=nd*3
        if nt=='N':
            nd=nd*4
    return nd


def degenerated_nn_melting(seq,salt_cc,oligo_cc):
    """Determine the melting temperature minimum and maximum for degenerated oligonucleotides
       Accept 3 parameters: oligonucleotide sequence, salt concentration (M), oligonucleotide concentration (M)""" 
    global meltmax,meltmin, enthmax,enthmin,entrmax,entrmin
    meltmax,meltmin=-1000,1000
    enthmax,enthmin=-1e9,1e9
    entrmax,entrmin=-1e9,1e9

    def degeneratedOligo(seq,dseq,salt_cc,oligo_cc):
        global meltmax,meltmin, enthmax,enthmin,entrmax,entrmin
        if len(seq)==len(dseq):
            melt,entr,enth=nn_melting(dseq,salt_cc,oligo_cc)
            meltmin,meltmax=min(melt,meltmin),max(melt,meltmax)
            enthmin,enthmax=min(enth,enthmin),max(enth,enthmax)
            entrmin,entrmax=min(entr,entrmin),max(entr,entrmax)
            return
        nt=seq[len(dseq)]
        for i,b in enumerate(['A','C','G','T']):
            if (IUPAC.find(nt)+1) & 2**i==2**i:
                dseq=dseq+b
                degeneratedOligo(seq,dseq,salt_cc,oligo_cc)
                dseq=dseq[0:-1]

    degeneratedOligo(seq,'',salt_cc,oligo_cc)
    return meltmin,meltmax,enthmin,enthmax,entrmin,entrmax

def nn_melting(seq,salt_cc,oligo_cc):
    '''Determine the melting temperature (C) of oligonucleotide sequence using the nearest-neighbor algorithm
       Accept 3 parameters: oligonucleotide sequence, salt concentration (M), oligonucleotide concentration (M)'''

    # parameters from Santalucia, 1998 (http://www.pubmedcentral.gov/articlerender.fcgi?tool=pubmed&pubmedid=9465037)
    dHo_santalucia1998={'AA':-7.9,'TT':-7.9,'AT':-7.2,'TA':-7.2,'CA':-8.5,'TG':-8.5,'GT':-8.4,'AC':-8.4,'CT':-7.8,'AG':-7.8,'GA':-8.2,'TC':-8.2,'CG':-10.6,'GC':-9.8,'GG':-8.0,'CC':-8.0}
    dSo_santalucia1998={'AA':-22.2,'TT':-22.2,'AT':-20.4,'TA':-21.3,'CA':-22.7,'TG':-22.7,'GT':-22.4,'AC':-22.4,'CT':-21.0,'AG':-21.0,'GA':-22.2,'TC':-22.2,'CG':-27.2,'GC':-24.4,'GG':-19.9,'CC':-19.9}    

    

    iupac_seq=clear_seq(convert2IUPAC(seq))

    if not iupac_seq:
       return '-'
    if len(iupac_seq) < 8:
        return 'Sequence too short', 0, 0
    if len(iupac_seq) > 50:
        return 'Sequence too long', 0, 0        
   
    #control if sequence is degenerated
    flag_degen = (ndegen(iupac_seq) > 1)

    if flag_degen:
        meltmin, meltmax, enthmin, enthmax, entrmin, entrmax = degenerated_nn_melting(iupac_seq, salt_cc, oligo_cc)
        return str(meltmin)+' - '+str(meltmax),str(enthmin)+' - '+str(enthmax),str(entrmin) + ' - ' + str(entrmax)
    else:
    
        enth,entr=0,0
        for i in range(len(iupac_seq)-1):
            enth=enth+dHo_santalucia1998[iupac_seq[i]+iupac_seq[i+1]]
            entr=entr+dSo_santalucia1998[iupac_seq[i]+iupac_seq[i+1]]
       
        if iupac_seq[0] in ['A','T']:
            enth+=2.3
            entr+=4.1
        
        if iupac_seq[0] in ['C','G']:
            enth+=0.1
            entr+=-2.8

        if iupac_seq[-1] in ['A','T']:
            enth+=2.3
            entr+=4.1
        
        if iupac_seq[-1] in ['C','G']:
            enth+=0.1
            entr+=-2.8

        enth=enth*4.18*1000
        entr=entr*4.18
    
        tm=enth/(entr+0.368*(len(iupac_seq)-1)*math.log(salt_cc)*4.18+1.987*4.18*math.log(oligo_cc/4))-273.15
    
    return round(tm, 2),round(enth,2),round(entr,2)

def MEC(seq,mecA=15400,mecC=7300,mecG=11700,mecT=8800):
    '''Determine the Molar extinction coefficient of nucleotide sequence'''
    mec=0
#    mecA=15400
#    mecC=7300
#    mecG=11700
#    mecT=8800
    ntMEC={'A':mecA,'C':mecC, 'G':mecG, 'T':mecT,'M':(mecA+mecC)/2,'R':(mecA+mecG)/2, 'S':(mecC+mecG)/2, 'V':(mecA+mecC+mecG)/3, 'W':(mecA+mecT)/2, 'Y':(mecC+mecT)/2, 'H':(mecA+mecC+mecT)/3, 'K':(mecG+mecT)/2, 'D':(mecA+mecG+mecT)/3, 'B':(mecC+mecG+mecT)/3, 'N':(mecA+mecC+mecG+mecT)/4}
    seq=clear_seq(convert2IUPAC(seq))
    for nt in seq:
        mec=mec+ntMEC[nt]
    return mec

def MW(seq):
    '''Determine the molecular weight of nucleotide sequence'''
    mw=0
    ntMW={'A':313.21,'C':289.19, 'G':329.21,'T':304.2, 'M':301.2,'R':321.21, 'S':309.2, 'V':310.54, 'W':308.7, 'Y':296.7, 'H':302.2, 'K':316.7, 'D':315.5, 'B':307.53, 'N':308.95}
    
    #print seq
    
    seq=clear_seq(seq)
    
    #print seq
    
    seq=convert2IUPAC(seq)
    
    #print seq
    
    for nt in seq:
        mw=mw+ntMW[nt]
    return mw

def GCp(seq):
    '''Determine the G+C content of nucleotide sequence'''
    count_gc,count_tot = 0,0
    seq = str(seq)
    seq = seq.upper()
    for nt in seq:
        if nt in ['C','G','S','N']:
            count_gc += 1
        if nt in ['M','R','Y','K']:
            count_gc += 0.5
        if nt in ['V','B']:
            count_gc += 0.667
        if nt in ['H','D']:
            count_gc += 0.333
        if nt in valid_nt:
            count_tot += 1
    if count_tot == 0:
        return 0.0
    else: 
        return round(count_gc*1.0/count_tot,2);

def clear_seq(seq):
    """remove all character except nucleotide"""
    seq2 = ''
    seq = str(seq)
    for nt in seq:
        if nt.upper() in valid_nt:
            seq2 = seq2 + nt
    return seq2

def GCG_checksum(seq):
    '''Determine the GCG checksum of nucleotide sequence'''
    seq=clear_seq(seq).upper()
    cs,count=0,0
    for nt in seq:
        if nt in valid_nt:
            count+=1
            cs=cs+count*ord(nt);
            if count==57:
                count=0
    cs=cs-(cs / 10000)*10000;
    return str(cs)

def seguid(seq):
    '''Determine the SEGUID checksum of nucleotide sequence'''
    try:
        import hashlib
        m = hashlib.sha1()
    except:
        import sha
        m = sha.new()
    import base64
    #print type(seq)
    if type(seq)!=type("aa"):
        seq=seq.tostring().upper()
    else:
        seq=seq.upper()
    m.update(seq)
    try:
        return base64.b64encode(m.digest()).rstrip("=")
    except:
        import os
        return base64.encodestring(m.digest()).replace(os.linesep,"").rstrip("=")

def seqChecksum(seq,format):
    if not format in ['SEGUID','GCG']:
        return 'Checksum type error'
    if format=='SEGUID':
        return seguid(seq)
    if format=='GCG':
        return GCG_checksum(seq)
    
    

def seq_translation(name,seq,frames,aa_code,flagIncludeNtSeq):
    '''translate sequence'''
    cg1=['GCT GCC GCA GCG',
         'TGT TGC',
         'GAT GAC',
         'GAA GAG',
         'TTT TTC',
         'GGT GGC GGA GGG',
         'CAT CAC',
         'ATT ATC ATA',
         'AAA AAG',
         'TTG TTA CTT CTC CTA CTG',
         'ATG',
         'AAT AAC',
         'CCT CCC CCA CCG',
         'CAA CAG',
         'CGT CGC CGA CGG AGA AGG',
         'TCT TCC TCA TCG AGT AGC',
         'ACT ACC ACA ACG',
         'GTT GTC GTA GTG',
         'TGG',
         'TAT TAC',
         'TAA TAG TGA']
    cg2=['Ala','Cys','Asp','Glu','Phe','Gly','His','Ile','Lys','Leu','Met','Asn','Pro','Gln','Arg','Ser','Thr','Val','Trp','Tyr',' * ']
    cg3=['A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R','S','T','V','W','Y','*']

    tr={1:'',2:'',3:'',4:'',5:'',6:''}
    r=''
    #sens frames
    for frame in [1,2,3]:
        if not frame in frames:
            continue

        t=frame-1

        while t<=len(seq)-3:
            codon=seq[t]+seq[t+1]+seq[t+2];
            flagCodonFound=False
            for t1 in range(0,21):

                if cg1[t1].find(codon)<>-1:
                    flagCodonFound=True

                    if aa_code==3:
                        tr[frame]=tr[frame]+cg2[t1]
                    if aa_code==1:
                        tr[frame]=tr[frame]+cg3[t1]+'  '
                    break

            if not flagCodonFound:
                if aa_code==3:
                    tr[frame]=tr[frame]+' ? '
                if aa_code==1:
                    tr[frame]=tr[frame]+'X'+'  '
                                                            
            t=t+3

    #antisens frames
    for frame in [4,5,6]:

        t=len(seq)+4-frame-1
        if not frame in frames:
            continue
        while t>=2:

            codon=complement_nt(seq[t])+complement_nt(seq[t-1])+complement_nt(seq[t-2])
            flagCodonFound=False
            for t1 in range(0,21):
                if cg1[t1].find(codon)<>-1:

                    flagCodonFound=True
                    if aa_code==3:
                       tr[frame]=cg2[t1]+tr[frame]
                    if aa_code==1:
                       tr[frame]='  '+cg3[t1]+tr[frame]
                    break


            if not flagCodonFound:
                if aa_code==3:
                    tr[frame]=tr[frame]+' ? '+tr[frame]
                if aa_code==1:
                    tr[frame]='  '+'X'+tr[frame]
               
            t=t-3


    if not flagIncludeNtSeq:
        for frame in frames:
            r=r+'>%s - frame %d\n' % (name,frame)
            r=r+tr[frame].replace(' ','')+'\n'

    else:
        t1=len(seq);
        if (t1 % 3)==0:
           tr[5]='  '+tr[5]
           tr[6]=' '+tr[6]

        if (t1 % 3)==1:
           tr[4]=' '+tr[4]
           tr[6]='  '+tr[6]

        if (t1 % 3)==2:
          tr[4]='  '+tr[4]
          tr[5]=' '+tr[5]

        t=1
    
        while seq:

            for frame in [1,2,3]:
                if frame in frames:

                    s3=tr[frame][0:60]
                    tr[frame]=tr[frame][60:]
                    r=r+'   '+str(frame)+':     '+'  '[0:frame-1]+s3+'\n'                    
    
            r=r+'\n'
    
            if flagIncludeNtSeq:
                #sense nt seq
                s3=seq[0:60]
                r=r+'          '+s3+'\n'

                #rules +---+
                s2=str(t)+' '
                while len(s2)<10:
                    s2=' '+s2
                
                for t1 in range(1,len(s3)+1):
                    if (t1 % 10)==0:
                        s2=s2+'+'
                    else:
                        s2=s2+'-'
                r=r+s2+'\n'
    
                #antisens nt seq
                s2=''
                for nt in s3:
                    s2=s2+complement_nt(nt);
                r=r+'          '+s2+'\n\n'

    
            for frame in [4,5,6]:
                if frame in frames:

                    s3=tr[frame][0:60]
                    tr[frame]=tr[frame][60:]
                    
                    r=r+'   '+str(frame)+':     '+s3+'\n'
    
            r=r+'\n\n'
            seq=seq[60:]
            t=t+60
    
    return r

def long_seq_melting(salt_cc,GCp,formamide_cc,seq_len):
    '''Determine the melting temperature of long (>50 nt) nucleotide sequence according to Howley, 1979'''

    return 81.5+16.6*math.log10(salt_cc)+0.41*GCp-0.62*formamide_cc-500/seq_len

def crop_sequence(seq,first,last):
    return seq[first-1:last]