#!/usr/bin/env python

"""local alignment of 2 nucleotidic sequences
by Smith and Waterman algorithm
"""

from __future__ import division

import numpy

def local_alignment(s1, s2):

    ### convert from QString to string
    s1 = str(s1)
    s2 = str(s2)

    output = []

    matchW = 10
    mismatchW = -9
    gapW = -20

    ni = len(s1)
    nj = len(s2)

    n = numpy.zeros((ni + 1, nj + 1), dtype=int)

    maxi = -1000000

    ### fill matrix
    for i in range(1, ni + 1):
        for j in range(1, nj + 1):
            if s1[i - 1] == s2[j - 1]:
                a = n[i - 1, j - 1] + matchW
            else:
                a = n[i - 1, j - 1] + mismatchW
            b = n[i, j - 1] + gapW
            c = n[i - 1, j] + gapW
            n[i, j] = max(0, a, b, c)


    ### screen matrix for all max scores
    maxi = n.max()

    if maxi == 0:
        print 'no hits'
        return {}

    for i2 in range(1, ni + 1):
        for j2 in range(1, nj + 1):

            if n[i2, j2] == maxi:

                mi = i2
                mj = j2
                ### start backtracking
                i = mi
                j = mj

                l1 = s1[i - 1]
                l2 = s2[j - 1]

                while True:

                    a = n[i - 1, j - 1]
                    b = n[i, j - 1]
                    c = n[i - 1, j]


                    if (i == 1 or j == 1 or n[i, j] == 0) or (a == 0 and b == 0 and c == 0):
                        l1 = ''.join([c for c in reversed(l1)])
                        l2 = ''.join([c for c in reversed(l2)])

                        rd = {}
                        ### position of seq2 on seq1

                        rd['init_alignment'] = i - j + 1

                        ### score
                        rd['score'] = maxi / min(ni, nj) / matchW * 100

                        ### query seq
                        query_seq = (' '*(i - j) + s2[1 - 1:j - 1] + l2 + s2[mj + 1 - 1:]).strip()
                        rd['aligned_query_seq'] = query_seq


                        k1 = ' ' * (j - i) + s1[1 - 1:i - 1] + l1 + s1[mi + 1 - 1:]


                        k2 = ' ' * (i - j) + s2[0:j - 1] + l2 + s2[mj + 1 - 1:]
                        if i - j + 1 - 1 < 0:
                            m = 0
                        else:
                            m = i - j + 1 - 1


                        k = k1[m:m + len(k2.strip())]

                        rd['aligned_subject_seq'] = k

                        align_len = len(s2)
                        if s2.count('-'):
                            align_len = len(query_seq)
                        cc = k.count('-')
                        if cc:
                            align_len = len(k) - cc

                        ### end alignment position 
                        rd['end_alignment'] = rd['init_alignment'] + align_len - 1

                        ### match
                        match = ''

                        for kk in range(0, min(len(k), len(query_seq))):

                            if query_seq[kk] == k[kk]:
                                match += '|'
                            else:
                                match += ' '

                        rd['match'] = match


                        output.append(rd)

                        break #continue for loop

                    if (n[i, j] == n[i - 1, j - 1] + matchW and s1[i - 1] == s2[j - 1]) or (n[i, j] == n[i - 1, j - 1] + mismatchW and s1[i - 1] <> s2[j - 1]):
                        i = i - 1
                        j = j - 1
                        l1 = l1 + s1[i - 1]
                        l2 = l2 + s2[j - 1]
                        continue  ### while

                    if n[i, j] == n[i, j - 1] + gapW:
                        j = j - 1
                        l1 = l1[0:-1] + '-' + l1[-1]
                        l2 = l2 + s2[j - 1]
                        continue    ### while

                    if n[i, j] == n[i - 1, j] + gapW:
                        i = i - 1
                        l1 = l1 + s1[i - 1]
                        l2 = l2[0:-1] + '-' + l2[-1]
                        continue    ### while

                    break

    return output


